# coding: utf-8

""" Parsing inputs into a payload. """

from __future__ import division, print_function

__author__ = "Andy Casey <andy@astrowizici.st>"

# Standard library
import datetime
import time
    
# Module specific
from utils import API_MAX_ROWS

__all__ = ['rows', 'ordering', 'dates', 'start']


def query(query, authors, dates):

    if query is None:
        # With great power comes great responsibility.
        return " *"

    # Regex match for a date range?

    # Assume rest is author/title interpreted?
    return query


def affiliation(affiliation):
    """Creates a parser based on the affiliation input."""

    if affiliation is None: return

    try:
        affiliation = str(affiliation)
    except TypeError:
        raise TypeError("affiliation must be a string-type")

    query_refinement = " aff:\"{affiliation}\"".format(affiliation=affiliation)
    return query_refinement


def rows(start, rows):
    """Checks that the number of rows provided is valid."""

    if rows == "all" or rows > API_MAX_ROWS:
        # We may have to run multiple queries here
        start, rows = 0, API_MAX_ROWS

    else:

        try:
            start = int(start)
        except TypeError:
            raise TypeError("start must be an integer-like type")

        if start < 0:
            raise ValueError("start must be positive")

        try: rows = int(rows)
        except TypeError:
            raise TypeError("rows must be an integer-like type")

        if rows < 1:
            raise ValueError("rows must be a positive integer")

    return start, rows


def ordering(sort, order):
    """Checks that the ordering and sorting inputs are valid."""

    sort, order = sort.lower(), order.lower()

    if order not in ('asc', 'desc'):
        raise ValueError("order must be either 'asc' or 'desc'")

    acceptable_sorts = {
        "date": "time",
        "relevance": "relevance",
        "cited": "citations",
        "popularity": "popular"
    }

    if sort not in acceptable_sorts \
    and sort not in acceptable_sorts.values():
        raise ValueError("sort must be one of: {acceptable_sorts}"
            .format(acceptable_sorts=', '.join(acceptable_sorts.keys())))

    if sort not in acceptable_sorts:
        for key in acceptable_sorts:
            if sort in acceptable_sorts[key]:
                sort = key
                break

    return (sort, order)


def _date(date_str, default_month=None, output_format="%Y-%m"):
    """Parses a date input into the preferred format for ADS."""

    if date_str in (None, "*"): return "*"

    date_str = str(date_str)

    formats = ("%Y", "%Y-%m", "%Y-%m-%d", "%Y/%m", "%Y/%m/%d")

    time_struct = None
    for i, format in enumerate(formats):
        try:
            time_struct = time.strptime(date_str, format)
        
        except ValueError:
            pass

        else:
            break

    if i == 0 and default_month is not None:
        time_struct = time.strptime("{year}-{month:.0f}"
            .format(year=time_struct.tm_year, month=default_month), "%Y-%m")
        
    if time_struct is None:
        raise ValueError("cannot parse date string '{date_str}'"
            .format(date_str=date_str))

    return time.strftime(output_format, time_struct)
    

def dates(input_dates):
    """Parses the input dates and returns a filter to append to the
    current query if the date range is restricted."""


    if input_dates is None: return

    if isinstance(input_dates, (list, tuple)):
        if len(input_dates) > 2:
            raise ValueError("if dates is specified as a list, it must"
                " be as '(start, end)', either of which can be empty")

        if len(input_dates) == 1:
            start_date, end_date = (input_dates[0], "*")

        else:
            start_date, end_date = input_dates

        start_date = _date(start_date, default_month=1)
        end_date = _date(end_date, default_month=12)

    elif isinstance(input_dates, (str, unicode)):

        # We will accept '2002-', '2002..', '..2003', '2002..2003'
        # '2002/01..2002/08', '2002/04..', '2002/07', '2002-7', '2002-07

        if input_dates.endswith("-") or input_dates.endswith(".."):
            start_date = _date(input_dates.strip("-."), default_month=1)
            end_date = "*"

        elif input_dates.startswith("-") or input_dates.startswith(".."):
            start_date = "*"
            end_date = _date(input_dates.strip("-."), default_month=12)

        elif ".." in input_dates:
            start_date, end_date = input_dates.split("..")

            if len(end_date) == 2:
                end_date = start_date[:2] + end_date

            start_date = _date(start_date, default_month=1)
            end_date = _date(end_date, default_month=12)

        else:
            if "/" in input_dates or "-" in input_dates:
                date = _date(input_dates)
            else:
                date = _date(int(input_dates), output_format="%Y")

            start_date, end_date = date, date

    elif isinstance(input_dates, (int, float)):

        # Is there a month component as a fractional?
        if input_dates % 1 > 0:
            date = _date("{year:.0f}/{month:.0f}"
                .format(year=int(input_dates), month=(input_dates % 1)*100))

        else:
            date = _date(int(input_dates), output_format="%Y")

        start_date, end_date = date, date

    else:
        raise TypeError("date type not understood")

    if "*" not in (start_date, end_date) and start_date > end_date:
        raise ValueError("start date ({start_date}) is after the end date ({end_date})"
            .format(start_date=start_date, end_date=end_date))

    # Build query
    query_refinement = " pubdate:[{start_date} TO {end_date}]".format(
        start_date=start_date, end_date=end_date)

    return query_refinement




# coding: utf-8

""" A Python module for NASA's Astrophysical Data Service (ADS) that doesn't suck. """

import os
import re
import sys

try:
    from setuptools import setup

except ImportError:
    from distutils.core import setup

major, minor1, minor2, release, serial =  sys.version_info

readfile_kwargs = {"encoding": "utf-8"} if major >= 3 else {}

def readfile(filename):
    with open(filename, **readfile_kwargs) as fp:
        contents = fp.read()
    return contents

version_regex = re.compile("__version__ = \"(.*?)\"")
contents = readfile(os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    "ads",
    "__init__.py"))

version = version_regex.findall(contents)[0]

setup(name="ads",
      version=version,
      author="Andrew R. Casey",
      author_email="andy@astrowizici.st",
      packages=["ads"],
      url="http://www.github.com/andycasey/ads-python/",
      license="GPLv2",
      description="A Python module for NASA's ADS that doesn't suck.",
      long_description=readfile(os.path.join(os.path.dirname(__file__), "README.md")),
      install_requires=[
        "requests",
        "requests_futures"
      ]
     )



# coding: utf-8

""" Searching for publications through NASA's ADS system. """

from __future__ import division, print_function

__author__ = "Andy Casey <andy@astrowizici.st>"

# Standard library
import json
import logging
import multiprocessing
import os
import time

# Third party
import requests
import requests_futures.sessions

# Module specific
import parser as parse
from utils import get_dev_key, get_api_settings, ADS_HOST, API_MAX_ROWS

__all__ = ["Article", "search", "metrics", "metadata", "retrieve_article"]

DEV_KEY = get_dev_key()

class Article(object):
    """An object to represent a single publication in NASA's Astrophysical
    Data System."""

    aff = ["Unknown"]
    author = ["Anonymous"]
    citation_count = 0
    url = None

    def __init__(self, **kwargs):
        for key, value in kwargs.iteritems():
            setattr(self, key, value)

        if "bibcode" in kwargs:
            self.url = "http://adsabs.harvard.edu/abs/{0}".format(kwargs["bibcode"])

        return None


    #__str__ should be readable
    def __str__(self):
        return u"<{0} {1} {2}, {3}>".format(self.author[0].split(",")[0],
            "" if len(self.author) == 1 else (u" & {0}".format(self.author[1].split(",")[0]) if len(self.author) == 2 else "et al."),
            self.year, self.bibcode)
    
    #__repr__ should be unambiguous
    def __repr__(self):
        return u"<ads.{0} object at {1}>".format(self.__class__.__name__, hex(id(self)))


    # TODO bibtex @property


    @property
    def references(self):
        """Retrieves reference list for the current article and stores them."""
        if not hasattr(self, '_references'):
            self._references = list(search("references(bibcode:{bibcode})".format(bibcode=self.bibcode), rows="all"))
        return self._references


    @property
    def citations(self):
        """Retrieves citation list for the current article and stores them."""
        if not hasattr(self, '_citations'):
            self._citations = list(search("citations(bibcode:{bibcode})".format(bibcode=self.bibcode), rows="all"))
        return self._citations


    @property
    def metrics(self):
        """Retrieves metrics for the current article and stores them."""

        if hasattr(self, "_metrics"):
            return self._metrics

        url = "{0}/record/{1}/metrics/".format(ADS_HOST, self.bibcode)
        payload = {"dev_key": DEV_KEY}

        r = requests.get(url, params=payload)
        if not r.ok: r.raise_for_status()

        self._metrics = r.json()
        return self._metrics


    def build_reference_tree(self, depth):
        """Builds a reference tree for this paper.

        Inputs
        ------
        depth : int
            The number of levels to fetch in the reference tree.

        Returns
        -------
        num_articles_in_tree : int
            The total number of referenced articles in the reference tree.
        """

        try: depth = int(depth)
        except TypeError:
            raise TypeError("depth must be an integer-like type")

        if depth < 1:
            raise ValueError("depth must be a positive integer")

        session = requests_futures.sessions.FuturesSession()

        # To understand recursion, first you must understand recursion.
        level = [self]
        total_articles = len(level) - 1

        for level_num in xrange(depth):

            level_requests = [search("references(bibcode:{bibcode})".format(bibcode=article.bibcode), rows="all") for article in level]

            # Complete all requests
            new_level = []
            for request, article in zip(level_requests, level):
                setattr(article, "references", list(request))
                new_level.extend(article.citations)

            level = sum([new_level], [])
            total_articles += len(level)

        return total_articles          


    def build_citation_tree(self, depth):
        """Builds a citation tree for this paper.

        Inputs
        ------
        depth : int
            The number of levels to fetch in the citation tree.

        Returns
        -------
        num_articles_in_tree : int
            The total number of cited articles in the citation tree.
        """

        try: depth = int(depth)
        except TypeError:
            raise TypeError("depth must be an integer-like type")

        if depth < 1:
            raise ValueError("depth must be a positive integer")

        session = requests_futures.sessions.FuturesSession()

        # To understand recursion, first you must understand recursion.
        level = [self]
        total_articles = len(level) - 1

        for level_num in xrange(depth):

            level_requests = [search("citations(bibcode:{bibcode})".format(bibcode=article.bibcode), rows="all") for article in level]

            # Complete all requests
            new_level = []
            for request, article in zip(level_requests, level):
                setattr(article, "_citations", list(request))
                new_level.extend(article.citations)

            level = sum([new_level], [])
            total_articles += len(level)

        return total_articles     


def _build_payload(query=None, authors=None, dates=None, affiliation=None, filter=None,
    fl=None, facet=None, sort="date", order="desc", start=0, rows=20):
    """Builds a dictionary payload for NASA's ADS based on the input criteria."""

    query = parse.query(query, authors, dates)

    # Check inputs
    start, rows = parse.rows(start, rows)
    sort, order = parse.ordering(sort, order)

    # Filters
    pubdate_filter = parse.dates(dates)
    affiliation_filter = parse.affiliation(affiliation)

    filters = (pubdate_filter, affiliation_filter)
    for query_filter in filters:
        if query_filter is not None:
            query += query_filter

    payload = {
        "q": query,
        "dev_key": DEV_KEY,
        "sort": "{sort} {order}".format(sort=sort.upper(), order=order),
        "start": start,
        "fmt": "json",
        "rows": rows,
    }
    additional_payload = {
        "fl": fl,
        "filter": filter,
        "facet": facet
    }
    payload.update(additional_payload)
        
    return payload


def metrics(author, metadata=False):
    """ Retrieves metrics for a given author query """

    payload = {
        "q": author,
        "dev_key": DEV_KEY,
    }
    r = requests.get(ADS_HOST + "/search/metrics/", params=payload)
    if not r.ok: r.raise_for_status()
    
    contents = r.json()
    metadata, results = contents["meta"], contents["results"]

    if metadata:
        return (results, metadata)
    return results


def metadata(query=None, authors=None, dates=None, affiliation=None, filter="database:astronomy"):
    """Search ADS for the given inputs and just return the metadata."""

    payload = _build_payload(**locals())
    payload["rows"] = 1 # It's meta-data, baby.
    r = requests.get(ADS_HOST + "/search/", params=payload)
    if not r.ok: r.raise_for_status()
    
    metadata = r.json()["meta"]
    return metadata



class search(object):
    """Search ADS and retrieve Article objects."""

    def __init__(self, query=None, authors=None, dates=None, affiliation=None, filter="database:astronomy",
        fl=None, facet=None, sort="date", order="desc", start=0, rows=20):
        
        arguments = locals().copy()
        del arguments["self"]

        self.payload = _build_payload(**arguments)
        session = requests_futures.sessions.FuturesSession()

        self.active_requests = [session.get(ADS_HOST + "/search/", params=self.payload)]
        self.retrieved_articles = []
        
        # Do we have to perform more queries?
        if rows == "all" or rows > API_MAX_ROWS:

            # Get metadata from serial request
            metadata_payload = self.payload.copy()
            metadata_payload["rows"] = 1
            r = requests.get(ADS_HOST + "/search/", params=metadata_payload)
            if not r.ok: r.raise_for_status()
            metadata = r.json()["meta"]

            # Are there enough rows such that we actually have to make more requests?
            if API_MAX_ROWS >= metadata["hits"]: return

            if rows == "all":
                num_additional_queries = int(metadata["hits"]/API_MAX_ROWS)
                if not metadata["hits"] % API_MAX_ROWS: num_additional_queries -= 1

            else:
                num_additional_queries = int(rows/API_MAX_ROWS)
                if not rows % API_MAX_ROWS: num_additional_queries -= 1

            # Initiate future requests
            for i in xrange(1, num_additional_queries + 1):
                # Update payload to start at new point
                self.payload["start"] = i * API_MAX_ROWS

                # Limit total number of rows if required
                if rows != "all" and (i + 1) * API_MAX_ROWS > rows:
                    self.payload["rows"] = rows - i * API_MAX_ROWS

                self.active_requests.append(session.get(ADS_HOST + "/search/", params=self.payload))

    def __iter__(self):
        return self

    def next(self):

        if len(self.active_requests) == 0 and len(self.retrieved_articles) == 0:
            raise StopIteration

        if len(self.retrieved_articles) == 0:
            active_request = self.active_requests.pop(0)
            response = active_request.result().json()

            self.retrieved_articles.extend([Article(**article_info) for article_info in response["results"]["docs"]])
            
        if len(self.retrieved_articles) == 0:
            raise StopIteration

        return self.retrieved_articles.pop(0)


def retrieve_article(article, output_filename, clobber=False):
    """Download the journal article (preferred) or pre-print version
    of the article provided, and save the PDF to disk.

    Inputs
    ------
    article : `Article` object
        The article to retrieve.

    output_filename : str
        The filename to save the article to.

    clobber : bool, optional
        Overwrite the filename if it already exists.
    """

    if os.path.exists(output_filename) and not clobber:
        raise IOError("output filename (\"{filename}\") exists and we've been "
            "asked not to clobber it.".format(filename=output_filename))

    # Get the ADS url
    ads_redirect_url = "http://adsabs.harvard.edu/cgi-bin/nph-data_query"
    arxiv_payload = {
        "bibcode": article.bibcode,
        "link_type": "PREPRINT",
        "db_key": "PRE"
    }
    article_payload = {
        "bibcode": article.bibcode,
        "link_type": "ARTICLE",
        "db_key": "AST"
    }
    
    # Let's try and download the article from the journal first
    article_r = requests.get(ads_redirect_url, params=article_payload)

    if not article_r.ok or "Requested scanned pages are not available" in article_r.text:

        # Use the arxiv payload
        arxiv_r = requests.get(ads_redirect_url, params=arxiv_payload)

        if not arxiv_r.ok:
            return False

        article_pdf_url = arxiv_r.url.replace("abs", "pdf")

    else:
        # Parse the PDF url
        article_pdf_url = None
    
    article_pdf_r = requests.get(article_pdf_url)
    if not article_pdf_r.ok: article_pdf_r.raise_for_status()

    with open(output_filename, "wb") as fp:
        fp.write(article_pdf_r.content)

    return True


